; ------------------------------------------------------------------
;  called-by-server.scm - Ben Vandenbos
;
;  *contains functions called by the server on the client's machine
; ------------------------------------------------------------------

; Called-by-Server Functions
; --------------------------

; NOTE: Functions called by the server cannot use the game-state semaphore
;       or call any other functions that use it, because they already own it.  

; used by server to report errors to client
(define error
  (lambda (str)
    (if *debug*
        (begin
          (display "ERROR reported by Server: ")
          (display str)
          (newline)))
    (disp-console (string-append "ERROR reported by Server: " str))))

; display a message on the client, from the server
(define chat
  (lambda (str)
    (disp-console str)))

; used by server to add objects to the client's game-state
(define add-object
  (lambda (index type x y theta movement rotate color owner)
    (display "Adding object...\n")
    (let ((time-stamp (current-milliseconds)))
      (if (>= index (vector-length *game-state*))
          (begin
            (set! *game-state* (vec-grow *game-state* (* 2 (vector-length *game-state*))))
            (add-object index type x y theta movement rotate color owner))
          (begin
            (cond
              ((eqv? type 'tank)
               (vector-set! *game-state* index (make-object type x y theta movement rotate color owner time-stamp 
                                                 (thread (lambda () (make-object-move index)))))
               #t)
              ((eqv? type 'bullet)
               (vector-set! *game-state* index (make-object type x y theta movement rotate color owner time-stamp 
                                                 (thread (lambda () (make-object-move index)))))
               #t)
              (#t #f)))
          )
      )
    )
  )

; used by the server to update object's in the client's game-state
(define update-object
  (lambda (index x y theta movement rotate color)
    (let ((time-stamp (current-milliseconds)))
      (if (>= index (vector-length *game-state*))
          (begin
            (if *debug* (display "Server attempted to update an out-of-bounds object\n"))
            (disp-console "Server attempted to update an out-of-bounds object."))
          (begin
            (set-object-x! (vector-ref *game-state* index) x)
            (set-object-y! (vector-ref *game-state* index) y)
            (set-object-theta! (vector-ref *game-state* index) theta)
            (set-object-movement! (vector-ref *game-state* index) movement)
            (set-object-rotate! (vector-ref *game-state* index) rotate)
            (set-object-color! (vector-ref *game-state* index) color)
            (set-object-time-stamp! (vector-ref *game-state* index) time-stamp)
            )
          )
      )
    )
  )

; used to hand shake with the client
(define hello
  (lambda (remote-game remote-version)
    (if (and (eqv? remote-game *game-name*)
             (eqv? remote-version *version*))
        (begin
          (send-command `(ok 'blue 1.0))
          (disp-console (string-append "Connected to " *server-address* ":" *server-port*))))
    )
  )

; used to set which index this client's tank is in
(define you-are
  (lambda (index)
    (set! *you-are-index* index)
    (set! *playing* #t)
    ; start a thread to dectect collisions between client's tank and everything else
    ;  it's ok to kill the collision thread, because we know this thread own the 
    ;  game-state semaphore
    (if (thread? *collision-thread*)
        (kill-thread *collision-thread*))
    (set! *collision-thread* (thread detect-collisions))
    (let ((x (object-x (vector-ref *game-state* *you-are-index*)))
          (y (object-y (vector-ref *game-state* *you-are-index*))))
      (set! x (- x *pan-thresh*))
      (set! y (- y *pan-thresh*))
      
      (if (> x (- *actual-field-width* *width*))
          (set! x (- *actual-field-width* *width*)))
      (if (< x *width*)
          (set! x 0))
      (if (> y (- *actual-field-height* *height*))
          (set! y (- *actual-field-height* *height*)))
      (if (< y *height*)
          (set! y 0))
      
      (set! *x-offset* x)
      (set! *y-offset* y)
      )
    )
  )

; used to delete an object from the game-state
(define remove-object
  (lambda (index)
    ; If the index isn't out of bounds, and the object is valid
    (display "Removing object...\n")
    (if (and (< index (vector-length *game-state*))
             (object? (vector-ref *game-state* index)))
        (begin
          ; If it's the client that's getting removed (i.e. you died)
          ; It's ok to kill the collision thread because we know
          ; it doesn't have the game-state semaphore, because
          ; this thread has it.
          (if (eqv? index *you-are-index*)
              (begin
                (set! *playing* #f)
                ))
          ; if the GUI thread is valid...
          (if (thread? (object-thread-desc (vector-ref *game-state* index)))
              (begin
                (display "Killing GUI thread...[remove-object]\n")
                (kill-thread (object-thread-desc (vector-ref *game-state* index)))
                ; Clear the screen/bitmap
                (if *wire-frame?*
                    (begin
                      (send mem-wire-dc set-brush *white-brush*)
                      (send mem-wire-dc set-pen *no-pen*)
                      (send mem-wire-dc draw-rectangle 0 0 1600 1200))
                    (begin
                      (send mem-dc draw-bitmap game-board 0 0))
                    )
                )
              )
          ; if it's a tank getting removed and we're using sprites, play the explosion animation
          (if (and (not *wire-frame?*) (eqv? (object-type (vector-ref *game-state* index)) 'tank))
              (let ((pos (exact-position (vector-ref *game-state* index))))
                (thread (lambda () (draw-animation (car pos) (cadr pos) 0))))
              )
          )
        )
    (vector-set! *game-state* index 0)
    )
  )
