; ------------------------------------------------------------------------
;  exact-position.scm - Ben VandenBos
;
;  *Functions and constants for determining the exact position of objects
;  *Time related functions
; ------------------------------------------------------------------------

; Constants
; ---------
; Constants for the different types of objects are stored in the hash table *move-constants*.  The
; key is the type ('tank, 'bullet, etc...) and the value is an "object-constants" struct, which is
; defined below.

(define-struct object-constants
               (speed rotation-time rotate-speed draw-sprite  draw-wire-frame))

(define *tank-constants* (make-object-constants
                          .15                 ; Tank move speed (pix/ms)
                          2000                ; Time for full tank rotation (ms)
                          (/ (* 2 *pi*) 2000)   ; Tank rotate speed (rads/ms)
                          draw-sprite-tank    ; Function for drawing a sprite tank
                          draw-wire-tank      ; Function for drawing a wire-frame tank
                          ))
(define *bullet-constants* (make-object-constants
                            .40   ; Bullet speed (pix/ms)
                            0   ; Full rotate time (not needed for bullets), (ms)
                            0   ; Rotate speed (not needed for bullets), (ms)
                            draw-sprite-bullet
                            draw-wire-bullet
                            ))
(define *move-constants* (make-hash-table))
(hash-table-put! *move-constants* 'tank *tank-constants*)
(hash-table-put! *move-constants* 'bullet *bullet-constants*)

; Time Functions
; --------------
; get the ellapsed time since "start-time" (using server-time)
(define get-ellapsed-time
  (lambda (start-time)
    (abs (- (abs (current-milliseconds))
            (abs start-time)))))

; Return the exact position of "object" at the current time (server time)
;  Doesn't wait for the semaphore, but should already have it.
(define exact-position
  (lambda (object)
    (if (object? object)     
        (let ((type (object-type object))
              (x (object-x object))
              (y (object-y object))
              (theta (object-theta object))
              (movement (object-movement object))
              (rotate (object-rotate object))
              (radius 0)
              (ellapsed-time 0))
          (let ((constants (hash-table-get *move-constants* type)))
            (cond
              ; FOR MOVING FORWARD
              ((eqv? movement 'up)
               (cond ((eqv? rotate 'right)
                      (begin
                        ; radius = (speed)(full-rotation-time)/2pi
                        (set! radius (/ (* (object-constants-speed constants) (object-constants-rotation-time constants)) (* 2 *pi*)))
                        (set! ellapsed-time (get-ellapsed-time (object-time-stamp object)))
                        (set! x (+ x (- (* radius (sin (+ theta (* (object-constants-rotate-speed constants) ellapsed-time))))
                                        (* radius (sin theta)))))
                        (set! y (- y (- (* radius (cos (+ theta (* (object-constants-rotate-speed constants) ellapsed-time))))
                                        (* radius (cos theta)))))                  
                        (set! theta (+ theta (* (object-constants-rotate-speed constants) ellapsed-time)))
                        )
                      )
                     ((eqv? rotate 'left)
                      (begin
                        ; radius = (speed)(full-rotation-time)/2pi
                        (set! radius (/ (* (object-constants-speed constants) (object-constants-rotation-time constants)) (* 2 *pi*)))
                        (set! ellapsed-time (get-ellapsed-time (object-time-stamp object)))
                        (set! x (- x (- (* radius (sin (- theta (* (object-constants-rotate-speed constants) ellapsed-time))))
                                        (* radius (sin theta)))))
                        (set! y (+ y (- (* radius (cos (- theta (* (object-constants-rotate-speed constants) ellapsed-time))))
                                        (* radius (cos theta)))))                  
                        (set! theta (- theta (* (object-constants-rotate-speed constants) ellapsed-time)))  
                        )
                      )
                     ((eqv? rotate 'stop)
                      (begin
                        (set! ellapsed-time (get-ellapsed-time (object-time-stamp object)))
                        (set! x (+ x (* (object-constants-speed constants) ellapsed-time (cos theta))))
                        (set! y (+ y (* (object-constants-speed constants) ellapsed-time (sin theta))))                   
                        )
                      )
                     )
               )
              
              ; FOR MOVING BACKWARD
              ((eqv? movement 'down)
               (cond ((eqv? rotate 'right)
                      (begin
                        ; radius = (speed)(full-rotation-time)/2pi
                        (set! radius (/ (* (object-constants-speed constants) (object-constants-rotation-time constants)) (* 2 *pi*)))
                        (set! ellapsed-time (get-ellapsed-time (object-time-stamp object)))
                        (set! x (- x (- (* radius (sin (+ theta (* (object-constants-rotate-speed constants) ellapsed-time))))
                                        (* radius (sin theta)))))
                        (set! y (+ y (- (* radius (cos (+ theta (* (object-constants-rotate-speed constants) ellapsed-time))))
                                        (* radius (cos theta)))))                  
                        (set! theta (+ theta (* (object-constants-rotate-speed constants) ellapsed-time)))                  
                        )
                      )
                     ((eqv? rotate 'left)
                      (begin
                        ; radius = (speed)(full-rotation-time)/2pi
                        (set! radius (/ (* (object-constants-speed constants) (object-constants-rotation-time constants)) (* 2 *pi*)))
                        (set! ellapsed-time (get-ellapsed-time (object-time-stamp object)))
                        (set! x (+ x (- (* radius (sin (- theta (* (object-constants-rotate-speed constants) ellapsed-time))))
                                        (* radius (sin theta)))))
                        (set! y (- y (- (* radius (cos (- theta (* (object-constants-rotate-speed constants) ellapsed-time))))
                                        (* radius (cos theta)))))                  
                        (set! theta (- theta (* (object-constants-rotate-speed constants) ellapsed-time)))  
                        )
                      )
                     ((eqv? rotate 'stop)
                      (begin
                        (set! ellapsed-time (get-ellapsed-time (object-time-stamp object)))                  
                        (set! x (- x (* (object-constants-speed constants) ellapsed-time (cos theta))))
                        (set! y (- y (* (object-constants-speed constants) ellapsed-time (sin theta))))
                        )
                      )
                     )
               )
              ; FOR MOVING LEFT
              ((eqv? rotate 'left)
               (set! ellapsed-time (get-ellapsed-time (object-time-stamp object)))
               (set! theta (- theta (* (object-constants-rotate-speed constants) ellapsed-time)))
               )
              ; FOR MOVING RIGHT
              ((eqv? rotate 'right)
               (set! ellapsed-time (get-ellapsed-time (object-time-stamp object)))
               (set! theta (+ theta (* (object-constants-rotate-speed constants) ellapsed-time)))
               )
              )
            
            ; if we're against a wall...
            (if (equal? (object-type object) 'tank)
                (begin
                  (if (< x (/ *tank-height-width* 2))
                      (set! x (/ *tank-height-width* 2)))
                  (if (> x (- *actual-field-width* (/ *tank-height-width* 2)))
                      (set! x (- *actual-field-width* (/ *tank-height-width* 2))))
                  (if (< y (/ *tank-height-width* 2))
                      (set! y (/ *tank-height-width* 2)))
                  (if (> y (- *actual-field-height* (/ *tank-height-width* 2)))
                      (set! y (- *actual-field-height* (/ *tank-height-width* 2))))))
            
            (list x y theta ellapsed-time)
            )
          )
        )
    )
  )
  